(***********************************************************************)
(*                                                                     *)
(*                          Sklml                                      *)
(*                                                                     *)
(*             Pierre Weis (Pierre.Weis@inria.fr)                      *)
(*             Francois Clement (Francois.Clement@inria.fr)            *)
(*             Quentin Carbonneaux (Quentin.Carbonneaux@inria.fr)      *)
(*                                                                     *)
(* This program is free software.                                      *)
(* (see the file LICENSE in the root directory of the distribution).   *)
(*                                                                     *)
(***********************************************************************)

(* $Id: sklstream_old.mli,v 1.29 2010-11-22 10:29:52 qcarbonn Exp $ *)

(** Streams are used to modelize the behaviour of computational nodes.

 Computational nodes are a kind of light servers or processes or threads, that get
 there input from the wires on the network and send their results on the wires.

 Since there is no thunk on the wires, there are no lazy values in the streams.
 Networks feature delay in values availability, hence there is a kind of lazy
 evaluation in networks that we modelize via lazy evaluation in the streams.

 In networks, even to get the first element of a stream, we have to explicitely
 ask and wait; hence a stream is a lazy data structure that has to be forced to get
 its head or even to know that it is indeed empty.
*)

(** The type that specifies what values are communicated on the wires,
    either real wires in the parallel interpretation, or on simulated wires
    in the sequential interpretation. *)

type 'a stream;;
(** The type of streams that run on the wires in parallel or
  are infinite data structures of lazy streams with ['a packet] elements.
*)

type 'a t = 'a stream;;
(** An alias for the [stream] type constructor. *)

(** Construction functions. *)
val singleton : 'a -> 'a stream;;
(** [singleton x] returns the stream that has only [x] as element. *)

(* exception End_of_stream;; *)
(** The exception raised by functions when a stream argument is empty. *)

val head : 'a stream -> 'a;;
(** [head s] returns the first element of stream [s].
  Raises [End_of_stream], if [s] is empty. *)

(* val unsingleton : 'a stream -> 'a;; *)

(** Generate a lazy stream from a function.
  The function must raise the exception [End_of_stream] when there is
  no more value to return. *)
(* val of_fun : (unit -> 'a) -> 'a stream;; *)

(** Generate a lazy stream from a function.
  The function must send options; the value [None] is sent when the stream
  must end. *)
val of_fun : (unit -> 'a option) -> 'a stream;;

(** Generate a (finite) lazy stream from a list. *)
val of_list : 'a list -> 'a stream;;

(** Generate a list from a lazy stream. *)
val to_list : 'a stream -> 'a list;;

(** Generate a (finite) lazy stream from a vector. *)
val of_vect : 'a array -> 'a stream;;

(** Generate the vector of the elements of a (finite) lazy stream. *)
val to_vect : 'a stream -> 'a array;;

val of_slist : 'a list stream -> 'a stream;;
val of_sarray : 'a array stream -> 'a stream;;

val to_slist : 'a stream -> 'a list stream;;
val to_sarray : 'a stream -> 'a array stream;;
(* val heads : 'a stream -> 'a stream;; *)

(** Operations on streams. *)

(** Merges two lazy streams into a single stream with intermixed elements from
    both streams in a fair way (fairly peeks up alternatively elements from
    both streams). *)
val join : ('a stream * 'a stream) -> 'a stream;;

(** Splits a stream into two. *)
val fork : 'a stream -> ('a stream * 'a stream);;

val loop : ('a stream -> bool stream) -> ('a stream -> 'a stream) ->
  'a stream -> 'a stream;;

val iter : ('a -> unit) -> 'a stream -> unit;;
val map : ('a -> 'b) -> 'a stream -> 'b stream;;
val foldl : ('a * 'b -> 'a) -> 'a -> 'b stream -> 'a stream;;

(** IO *)

val of_to_channel : (IO.ic * IO.oc) -> 'a stream -> 'b stream;;
val to_channel : 'a stream -> IO.oc -> unit;;
val of_channel : IO.ic -> 'a stream;;
val stream_to_channel : 'a stream -> IO.oc -> unit stream;;

(* DEBUGGING *)

val print_stream :
  (Format.formatter -> 'a -> unit) ->
  Format.formatter -> 'a stream -> unit
;;

val eq_stream : 'a stream -> 'a stream -> bool
;;
